# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
# SPDX-FileCopyrightText: 2021-2022 Harald Sitter <sitter@kde.org>

import os
import sys

sys.path.append(f'{os.path.dirname(os.path.dirname(os.path.abspath(__file__)))}/')

# Initialize sentry reports for exceptions in this script
# NOTE: this happens before other imports so we get reports when we have systems with missing deps
try:
    import sentry_sdk
    sentry_sdk.init(
        dsn="https://d6d53bb0121041dd97f59e29051a1781@crash-reports.kde.org/13",
        traces_sample_rate=1.0,
        release="drkonqi@" + os.getenv('DRKONQI_VERSION'),
        dist=os.getenv('DRKONQI_DISTRIBUTION'),
        ignore_errors=[KeyboardInterrupt],
        # Shutdown performance doesn't matter all that much, give us ample time to send a possible report instead.
        shutdown_timeout=30,
    )
except ImportError:
    print("python sentry-sdk not installed :(")

os.environ['LC_ALL'] = 'C.UTF-8'

import gdb
from gdb.FrameDecorator import FrameDecorator

from datetime import datetime
import uuid
import json
import subprocess
import signal
import re
import binascii
import platform
import multiprocessing
from pathlib import Path
import psutil

crashed_thread = None

class UnexpectedMappingException(Exception):
    pass

class NoBuildIdException(Exception):
    pass

def mangle_path(path):
    if not path:
        return path
    return re.sub(str(Path.home()), "$HOME", path, count=1)

class SentryQMLThread:
    def __init__(self):
        self.payload = None

        # TODO this is largely a code dupe of print_qml_trace

        if gdb.selected_inferior().connection.type == 'core':
            # Only live processes can be traced unfortunately since we need to
            # call a function on the process. That does not work on cores.
            return

        # should we iterate the inferiors? Probably makes no diff for 99% of apps.
        for thread in gdb.selected_inferior().threads():
            if not thread.is_valid() :
                continue
            thread.switch()
            if gdb.selected_thread() != thread:
                continue # failed to switch :shrug:

            try:
                frame = gdb.newest_frame()
            except gdb.error:
                pass
            while frame:
                ret = qml_trace_frame(frame)
                if ret:
                    self.payload = ret
                    break
                try:
                    frame = frame.older()
                except gdb.error:
                    pass

    def to_sentry_frame(self, frame):
        print(frame)
        blob = {
            'platform': 'other', # always different from the cpp/native frames. alas, technically this frame isn't a javascript frame
            'in_app': True # qml is always in the app I should think
        }
        if 'file' in frame: blob['filename'] = mangle_path(frame['file'])
        if 'func' in frame: blob['function'] = frame['func']
        if 'line' in frame: blob['lineno'] = int(frame['line'])
        return blob

    def to_sentry_frames(self, frames):
        lst = []
        for frame in frames:
            data = self.to_sentry_frame(frame)
            if not data:
                continue
            lst.append(data)
        return lst

    def to_dict(self):
        if not self.payload:
            return None

        payload = self.payload

        from pygdbmi import gdbmiparser
        result = gdbmiparser.parse_response("*stopped," + payload)
        frames = result['payload']['frame']
        print(frames)
        if type(frames) is dict: # single frames traces aren't arrays to make it more fun -.-
            frames = [frames]
        lst = self.to_sentry_frames(frames)
        print(lst)
        if lst:
            return {
                'id': 'QML', # docs say this is typically a number to there is indeed no enforcement it seems
                'name': 'QML',
                'crashed': True,
                'stacktrace': {
                    'frames': self.to_sentry_frames(frames)
                }
            }
        return None

    def to_list(self):
        data = self.to_dict()
        if data:
            return [data]
        return []

class SentryVariablesStatistics:
    # All seen frames
    frames_count: int = 1 # we divide by this, it had better be >0
    # Amount of frames that had local variables
    frames_with_vars: int = 0

    # Only ever consider frames in the crashing thread!
    # The other threads are 99% of the time not interesting.

    def increment_frames():
        if gdb.selected_thread() and gdb.selected_thread() == crashed_thread:
            SentryVariablesStatistics.frames_count = SentryVariablesStatistics.frames_count + 1

    def increment_vars():
        if gdb.selected_thread() and gdb.selected_thread() == crashed_thread:
            SentryVariablesStatistics.frames_with_vars = SentryVariablesStatistics.frames_with_vars + 1

    def vars_rate():
        return round(SentryVariablesStatistics.frames_with_vars / SentryVariablesStatistics.frames_count, 2)

# Only grabing the most local block, technically we could also gather up encompassing scopes but it may be a bit much.
class SentryVariables:
    def __init__(self, frame):
        self.frame = frame
        SentryVariablesStatistics.increment_frames()

    def block(self):
        try:
            return self.frame.block()
        except:
            return None

    def to_dict(self):
        ret = self._to_dict_internal()
        if len(ret) > 0:
            SentryVariablesStatistics.increment_vars()
        return ret

    def _to_dict_internal(self):
        ret = {}
        block = self.block()
        if not block:
            return ret

        for symbol in block:
            try:
                ret[str(symbol)] = str(symbol.value(self.frame))
            except:
                pass # either not a variable or not stringable
        return ret

class SentryFrame:
    def __init__(self, gdb_frame):
        self.frame = gdb_frame
        self.sal = gdb_frame.find_sal()

    def type(self):
        return self.frame.type()

    def filename(self):
        return self.sal.symtab.fullname() if (self.sal and self.sal.symtab) else None

    def lineNumber(self):
        if self.sal.line < 0:
            return None
        # NOTE "The line number of the call, starting at 1." - I'm almost sure gdb starts at 0, so add 1
        return self.sal.line + 1

    def function(self):
        return self.frame.name() or self.frame.function() or None

    def package(self):
        name = gdb.solib_name(self.frame.pc())
        if not name:
            return name
        # NOTE: realpath because neon's gdb is confused over UsrMerge symlinking of /lib to /usr/lib messing up
        # path consistency (mapping data and by extension SentryImage instances use the real path already though
        return os.path.realpath(name)

    def address(self):
        return ('0x%x' % self.frame.pc())

    def to_dict(self):
        return {
            'filename': mangle_path(self.filename()),
            'function': self.function(),
            'package': mangle_path(self.package()),
            'instruction_addr': self.address(),
            'lineno': self.lineNumber(),
            'vars': SentryVariables(self.frame).to_dict()
        }

class SentryRegisters:
    def __init__(self, gdb_frame):
        self.frame = gdb_frame

    def to_dict(self):
        js = {}
        try: # registers() is only available in somewhat new gdbs. (e.g. not ubuntu 20.04)
            for register in self.frame.architecture().registers():
                if register.startswith('ymm'): # ymm actually contains stuff sentry cannot handle. alas :(
                    continue
                value = self.frame.read_register(register).format_string(format='x')
                if value: # may be empty if the value cannot be expressed as hex (happens for extra gdb register magic - 'ymm0' etc)
                    js[register.name] = value
                else:
                    js[register.name] = "0x0"
        except AttributeError:
            return None
        return js

class LockReason:
    def __init__(self, frame: SentryFrame, type, class_name):
        self.type = type
        self.class_name = class_name
        self.thread_id = gdb.selected_thread().ptid[1]

    def to_dict(self):
        return {
            'type': self.type,
            'class_name': self.class_name,
            'thread_id': self.thread_id,
            'package_name': 'java.lang',
        }

    def make(frame):
        # export enum LockType {
        #   LOCKED = 1,
        #   WAITING = 2,
        #   SLEEPING = 4,
        #   BLOCKED = 8,
        # }
        func = frame.function()
        match func:
            case 'QtLinuxFutex::_q_futex':
                return LockReason(frame, 8, 'QtLinuxFutex')
            case ('___pthread_cond_wait', 'pthread_cond_wait'):
                return LockReason(frame, 2, 'pthread_cond_wait')
            case 'QWaitCondition::wait':
                return LockReason(frame, 2, 'QWaitCondition')
        return None

class SentryTrace:
    def __init__(self, thread, is_crashed):
        thread.switch()
        self.frame = gdb.newest_frame()
        self.is_crashed = is_crashed
        self.lock_reasons = {}
        self.was_main_thread = False
        self.crashed = self.is_crashed # different from is_crashed (=input) this indicates if we stumbled over the kcrash handler

    def to_dict(self):
        frames = [ SentryFrame(frame) for frame in gdb.FrameIterator.FrameIterator(self.frame) ]
        self.lock_reasons = {}
        self.was_main_thread = False

        kcrash_index = -1
        trap_index = -1
        for index, frame in enumerate(frames):
            if frame.function():
                lock_reason = LockReason.make(frame)
                if lock_reason:
                    r = lock_reason.to_dict()
                    address = f'0x{str(len(self.lock_reasons.keys()))}'
                    r['address'] = address
                    self.lock_reasons[address] = r
                if frame.function().startswith('KCrash::defaultCrashHandler'):
                    kcrash_index = index
                    self.crashed = True
                if frame.function().startswith('QCoreApplication::exec'):
                    self.was_main_thread = True
            if frame.type() == gdb.SIGTRAMP_FRAME:
                trap_index = index
        clip_index = max(kcrash_index, trap_index)

        # Throw away kcrash or sigtrap frame, and above. They are useless noise - but only when on the crashing thread.
        if self.is_crashed and clip_index > -1:
            frames = frames[(clip_index + 1):]

        # Sentry format wants oldest frame first.
        frames.reverse()
        return { 'frames': [ frame.to_dict() for frame in frames ], 'registers': SentryRegisters(self.frame).to_dict() }

class SentryThread:
    def __init__(self, gdb_thread, is_crashed):
        self.thread = gdb_thread
        self.is_crashed = is_crashed

    def to_dict(self):
        # https://develop.sentry.dev/sdk/event-payloads/threads/
        # As per Sentry policy, the thread that crashed with an exception should not have a stack trace,
        #  but instead, the thread_id attribute should be set on the exception and Sentry will connect the two.
        trace = SentryTrace(self.thread, self.is_crashed)
        # NB: trace.to_dict creates members as side effect, run it asap
        payload = {
            'stacktrace': trace.to_dict(),
            'id': self.thread.ptid[1],
            'name': self.thread.name,
            'current': self.is_crashed,
            'crashed': trace.crashed, # side effect
            'main': trace.was_main_thread, # side effect
            'held_locks': trace.lock_reasons, # side effect
        }
        # States appear not documented. They are
        #   RUNNABLE = 'Runnable',
        #   TIMED_WAITING = 'Timed waiting',
        #   BLOCKED = 'Blocked',
        #   WAITING = 'Waiting',
        #   NEW = 'New',
        #   TERMINATED = 'Terminated',
        state = None
        if self.thread.is_exited():
            state = 'Terminated'
        if not state:
            for addr, reason in trace.lock_reasons.items():
                match reason['type']:
                    case 1:
                        state = None # locked doesn't exist as thread state
                    case 2:
                        state = 'Waiting'
                    case 4:
                        state = 'Runnable'
                    case 8:
                        state = 'Blocked'
                break
        payload['state'] = state
        return payload

class SentryImage:
    # NOTE: realpath hacks because neon's gdb is confused over UsrMerge symlinking of /lib to /usr/lib messing up
    # path consistency so always force realpathing for our purposes (this also is applied in SentryFrame)
    _objfiles = {}

    def objfiles(self):
        if SentryImage._objfiles:
            return SentryImage._objfiles

        objfiles = {}
        for objfile in gdb.objfiles():
            objfiles[objfile.filename] = objfile
            objfiles[os.path.realpath(objfile.filename)] = objfile
        SentryImage._objfiles = objfiles
        return objfiles

    # This can throw if objfiles fail to resolve!
    def __init__(self, file, start, end):
        # Awkwardly gdb python doesn't really give access to the solibs, meanwhile
        # the CLI doesn't really give access to the build_id. So we need to tuck
        # the two together to get comprehensive data on the loaded images.
        self.valid = False
        self.file = os.path.realpath(file)
        self.image_start = start
        self.image_end = end
        # Required! We can't build a debug_id without it and we require a debug_id!
        try:
            # If the mapped file isn't actually a library it will not be in the objfile rendering the image moot.
            # This happens because we need to construct off of proc mapping data. This also includes /dev nodes,
            # cache files and the like. The easiest way to filter them out is to check if the file is in the objfiles.
            self.objfile = self.objfiles()[self.file]
        except KeyError:
            if self.file.endswith(".so"):
                try:
                    lookup = gdb.lookup_objfile(self.file)
                except ValueError:
                    lookup = None
                objfiles = gdb.objfiles()
                self_objfiles = self.objfiles() # pull into scope so we have it in the trace in sentry
                if 'sentry_sdk' in globals():
                    progspace = gdb.selected_inferior().progspace
                    pid_running = psutil.pid_exists(gdb.selected_inferior().pid)
                    sentry_sdk.add_breadcrumb(
                        category='debug',
                        level='debug',
                        message=f'Progspace {progspace} :: {progspace.filename} :: {progspace.is_valid()} :: pid running ({pid_running})',
                    )
                    sentry_sdk.capture_exception(UnexpectedMappingException("unexpected mapping fail {} {} {} {}"
                                                                            .format(self.file, lookup, objfiles, self_objfiles)))
            return
        self.valid = True

    def debug_id(self):
        # Identifier of the dynamic library or executable.
        # It is the value of the build_id custom section and must be formatted
        # as UUID truncated to the leading 16 bytes.
        build_id = self.build_id()
        if not build_id:
            raise NoBuildIdException(f'Unexpectedly stumbled over an objfile ({self.file}) without build_id. Not creating payload.')
        truncate_bytes = 16
        build_id = build_id + ("00" * truncate_bytes)
        return str(uuid.UUID(bytes_le=binascii.unhexlify(build_id)[:truncate_bytes]))

    def build_id(self):
        return self.objfile.build_id

    def to_dict(self):
        if not self.valid:
            return None
        # https://develop.sentry.dev/sdk/event-payloads/debugmeta

        return {
            'type': 'elf',
            'image_addr': hex(self.image_start),
            'image_size': (self.image_end - self.image_start),
            'debug_id': self.debug_id(),
            # 'debug_file': None, # technically could get this from objfiles somehow but probably not useful cause it can't be used for anything
            'code_id': self.build_id(),
            'code_file': self.file,
            # 'image_vmaddr': None, # not available we'd have to read the ELF I think
            'arch': platform.machine(),
        }

def get_stdout(proc):
    proc = subprocess.run(proc, stdout=subprocess.PIPE)
    if proc.returncode != 0:
        return ''
    return proc.stdout.decode("utf-8").strip()

class SentryImages:
    _mapping_re = re.compile(
        r"""(?x)

        \s*

        (?P<start>
            0[xX][a-fA-F0-9]+
        )

        \s+

        (?P<end>
            0[xX][a-fA-F0-9]+
        )

        \s+

        (?P<size>
            0[xX][a-fA-F0-9]+
        )

        \s+

        (?P<offset>
            0[xX][a-fA-F0-9]+
        )

        \s+

        (
            (?P<permissions>
            [rwxps-]+)
            \s+
        )?

        (?P<file>
            [\/|\/][\w|\S]+|\S+\.\S+|[a-zA-Z]*
        )
        """
    )

    def __init__(self):
        # NB: gdb also has `info sharedlibrary` but that refers to section addresses inside the image. this would mess
        # up symbolication as we need the correct image start in the memory region. The only way to get that is through
        # proc mappings.
        mapping = {}
        try:
            output = gdb.execute('info proc mappings', to_string=True)
        except:
            return
        for line in output.splitlines():
            match = SentryImages._mapping_re.match(line)
            if not match:
                continue
            start = int(match.group('start'), 0)
            end = int(match.group('end'), 0)
            # we'll calculate size ourselves; the match is not used
            # offset basically just skips over previous sections so we don't really care
            file = match.group('file')
            if file not in mapping:
                mapping[file] = {'start': start, 'end': end}
                continue
            mapping[file]['start'] = min(mapping[file]['start'], start)
            mapping[file]['end'] = max(mapping[file]['end'], end)

        # TODO: if the regexing fails we could fall back to reading /proc/1/maps instead, I'd rather have more code than useless traces because of missing images
        self.mappings = mapping

    def to_list(self):
        ret = []
        if not self.mappings: return ret
        for file, mapping in self.mappings.items():
            image = SentryImage(file=file, start=mapping['start'], end=mapping['end'])
            if not image.valid: # images are invalid if the file wasn't actually found in the gdb.objfiles
                continue
            ret.append(image.to_dict())
        return ret

class SentryEvent:
    def cpu_model(self):
        with open("/proc/cpuinfo") as f:
            for line in f.readlines():
                key, value = line.split(':', 2)
                if key.strip() == 'model name':
                    return value.strip()
        return None

    def make(self, program, crash_thread):
        crash_signal = int(os.getenv('DRKONQI_SIGNAL'))
        vm = psutil.virtual_memory()
        boot_time = datetime.utcfromtimestamp(psutil.boot_time()).strftime('%Y-%m-%dT%H:%M:%S')

        # crutch to get the build id. if we did this outside gdb I expect it'd be neater
        progfile = gdb.current_progspace().filename
        build_id = gdb.lookup_objfile(progfile).build_id

        base_data = json.loads(get_stdout(['drkonqi-sentry-data']))
        sentry_event = { # https://develop.sentry.dev/sdk/event-payloads/
            "debug_meta": { # https://develop.sentry.dev/sdk/event-payloads/debugmeta/
                "images": SentryImages().to_list()
            },
            'threads':  [ # https://develop.sentry.dev/sdk/event-payloads/threads/
                 SentryThread(thread, is_crashed=(thread == crash_thread)).to_dict() for thread in gdb.selected_inferior().threads()
            ] # + SentryQMLThread().to_list(), TODO make qml more efficient it iterates everything again after the sentry threads were collected. a right waste of time!
            ,
            'event_id': uuid.uuid4().hex,
            # Gets overwritten by ReportInterface with a more accurate value
            'timestamp': datetime.utcnow().isoformat(),
            'message': 'Signal {} in {}'.format(crash_signal, program),
            'platform': 'native',
            'sdk': {
                'name': 'kde.drkonqi.gdb',
                'version': os.getenv('DRKONQI_VERSION'),
             },
            'level': 'fatal',
            # FIXME this is kind of wrong, program ought to be mapped to the project name via our DSNs mapping table (see reportinterface.cpp)
            'release': "{}@unknown".format(program),
            'dist': build_id,
            # TODO environment entry (could be staging for beta releases?)
            'contexts': { # https://develop.sentry.dev/sdk/event-payloads/contexts/
                'device': {
                    'name': base_data['Hostname'],
                    'model': self.cpu_model(),
                    'family': base_data['Chassis'],
                    'simulator': base_data['Virtualization'],
                    'arch': platform.machine(),
                    'memory_size': vm.total,
                    'free_memory': vm.available,
                    'boot_time': boot_time,
                    'timezone': base_data['Timezone'],
                    'processor_count': multiprocessing.cpu_count()
                },
                # 'os' gets injected on the cpp side so it is always available
            },
            'exception': { # https://develop.sentry.dev/sdk/event-payloads/exception/
                'values': [
                    {
                        'value': signal.strsignal(crash_signal),
                        'thread_id': crash_thread.ptid[1],
                        'mechanism': {
                            'type': 'drkonqi',
                            'handled': False,
                            "synthetic": True, # Docs: This flag should be set for all "segfaults"
                            'meta': {
                                'signal': {
                                    'number': crash_signal,
                                    'name': signal.strsignal(crash_signal)
                                },
                            },
                        },
                        'stacktrace': SentryTrace(crash_thread, True).to_dict(),
                    }
                ]
            },
        }

        # WARNING: must be after the trace is constructed so SentryVariables are counted
        sentry_event['tags'] = {
            'binary': program, # for fallthrough we still need a convenient way to identify things
            'stack_vars': 'yes' if (SentryVariablesStatistics.vars_rate() > 0.25) else 'no',
            'stack_vars_rate': str(SentryVariablesStatistics.vars_rate()), # must be str for sentry to consume it properly
        }

        if os.getenv('DRKONQI_APP_VERSION'):
            sentry_event['release'] = '{}@{}'.format(program, os.getenv('DRKONQI_APP_VERSION'))

        return sentry_event

def qml_trace_frame(frame):
    # NB: Super inspired by QtCreator's gdbbridge.py (GPL3).
    # I've made the code less of an eye sore though.

    # This is a very exhaustive attempt at finding a frame that has a symbol to the
    # QV4::ExecutionEngine as we need its address to get the QML trace via qt_v4StackTraceForEngine.
    # Unfortunately there's no shorter way of accomplishing this since the engine isn't necessarily
    # appearing as a frame (consequently we can't easily get to a this pointer).

    try:
        block = frame.block()
    except:
        block = None

    if not block:
        return None

    for symbol in block:
        if not symbol.is_variable and not symbol.is_argument:
            continue

        value = symbol.value(frame)
        if value.is_optimized_out: # can't read values that have been optimized out
            continue

        typeobj = value.type
        if typeobj.code != gdb.TYPE_CODE_PTR:
            continue

        dereferenced_type = typeobj.target().unqualified()
        if dereferenced_type.name != 'QV4::ExecutionEngine':
            continue

        addr = int(value)
        methods = [
            'qt_v4StackTraceForEngine((void*)0x{0:x})',
            'qt_v4StackTrace(((QV4::ExecutionEngine *)0x{0:x})->currentContext())',
            'qt_v4StackTrace(((QV4::ExecutionEngine *)0x{0:x})->currentContext)',
        ]
        for method in methods:
            try: # throws when the function is invalid
                result = str(gdb.parse_and_eval(method.format(addr))).strip()
            except gdb.error:
                continue
            if result:
                # We need to massage the result a bit. It's of the form
                #   "$addr    stack=[...."
                # but we want to drop the addr as it's not useful data and can't get parsed.
                # Also drop the stack nesting. Serves no purpose for us. Also unescape the quotes.
                pos = result.find('"stack=[')
                if pos != -1:
                    result = result[pos + 8:-2]
                    result = result.replace('\\\"', '\"')
                    return result

    return None

def print_qml_frame(frame):
    data = {
        'level': '?',
        'func': '?',
        'file': '?',
        'line': '?'
    }
    data.update(frame)
    print("level={level} func={func} at={file}:{line}".format(**data) )

def print_qml_frames(payload):
    from pygdbmi import gdbmiparser
    response = gdbmiparser.parse_response("*stopped," + payload)
    frames = response['payload']['frame']
    if type(frames) is dict: # single frames traces aren't arrays to make it more fun -.-
        print_qml_frame(frames)
    else: # presumably an iterable
        for frame in frames:
            print_qml_frame(frame)


def print_qml_trace():
    if gdb.selected_inferior().connection.type == 'core':
        # Only live processes can be traced unfortunately since we need to
        # call a function on the process. That does not work on cores.
        print('Cannot QML trace cores :(')
        return

    try:
        from pygdbmi import gdbmiparser
    except ImportError:
        print('Cannot QML trace cores because pygdbmi is missing :(')
        return

    # should we iterate the inferiors? Probably makes no diff for 99% of apps.
    for thread in gdb.selected_inferior().threads():
        if not thread.is_valid():
            continue
        thread.switch()
        if gdb.selected_thread() != thread:
            continue # failed to switch :shrug:

        try:
            frame = gdb.newest_frame()
        except gdb.error:
            pass
        while frame:
            ret = qml_trace_frame(frame)
            if ret:
                header = "____drkonqi_qmltrace_thread:{}____".format(str(thread.num))
                print(frame)
                print(header)
                print_qml_frames(ret)
                print('-' * len(header))
                print("(beware that frames may have been optimized out)")
                print() # separator newline
                break # next thread (there should only be one engine per thread I think?)
            try:
                frame = frame.older()
            except gdb.error:
                pass

def print_kcrash_error_message():
    symbol = gdb.lookup_static_symbol("s_kcrashErrorMessage")
    if not symbol or not symbol.is_valid():
        return

    try:
        value = symbol.value()
    except: # docs say value can throw!
        return
    print("KCRASH_INFO_MESSAGE: Content of s_kcrashErrorMessage: " + value.format_string())
    print() # separator newline

def print_sentry_payload(thread):
    program = os.path.basename(gdb.current_progspace().filename)
    payload = SentryEvent().make(program, thread)

    tmpdir = os.getenv('DRKONQI_TMP_DIR')
    if tmpdir:
        with open(tmpdir + '/sentry_payload.json', mode='w') as tmpfile:
            tmpfile.write(json.dumps(payload))
            tmpfile.flush()

def print_preamble_internal():
    thread = gdb.selected_thread()
    if thread == None:
        # Can happen when e.g. the core is missing or not readable etc. We basically aren't debugging anything
        return
    if 'sentry_sdk' in globals():
        sentry_sdk.add_breadcrumb(
            category='debug',
            level='debug',
            message=f'Selected thread {thread}',
        )
    global crashed_thread
    crashed_thread = thread
    # run this first as it expects the current frame to be the crashing one and qml tracing changes the frames around
    print_kcrash_error_message()
    # changes current frame and thread!
    print_qml_trace()
    # prints sentry report
    try:
        print_sentry_payload(thread)
    except NoBuildIdException as e:
        print(e)
        pass

def print_preamble():
    try:
        print_preamble_internal()
    except Exception as e:
        if 'sentry_sdk' in globals():
            sentry_sdk.capture_exception(e)
        raise(e)



# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
# SPDX-FileCopyrightText: 2023 Harald Sitter <sitter@kde.org>

import unittest
from unittest.mock import MagicMock, patch
from chai import Chai
import sys
import os
from pathlib import Path

os.environ['DRKONQI_VERSION'] = '1.2.3'

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(f'{SCRIPT_DIR}/python/')
sys.path.append(f'{os.path.dirname(SCRIPT_DIR)}/data/')

import gdb_preamble.preamble as preamble
import gdb

class PreambleTest(Chai):
    @classmethod
    def setUpClass(self):
        pass

    def setUp(self):
        preamble.SentryImage._objfiles = {}
        super(PreambleTest, self).setUp()

    def frame(self):
        self.mock(gdb, 'solib_name')
        self.expect(gdb.solib_name).args(0x1).returns(
            f'{Path.home()}/foo.so').at_least(0)

        symbol = MagicMock()
        symbol.__str__.return_value = 'i'
        symbol.value.return_value = 123

        symtab = MagicMock()
        symtab.fullname = MagicMock(return_value=f'{Path.home()}/foo.so')

        sal = MagicMock()
        sal.symtab = symtab
        sal.line = 0

        rax = self.mock(MagicMock)
        rax.name = "rax"
        self.expect(rax, 'startswith').args('ymm').returns(False).at_least(0)

        ymm0 = self.mock(MagicMock)
        ymm0.name = "ymm0"
        self.expect(ymm0, 'startswith').args('ymm').returns(True).at_least(0)

        architecture = self.mock()
        self.expect(architecture, 'registers').returns(['ymm0', rax]).at_least(0)

        rax_value = self.mock()
        self.expect(rax_value, 'format_string').args(format='x').returns('0x2').at_least(0)

        f = self.mock()
        self.expect(f, 'block').returns([symbol]).at_least(0)
        self.expect(f, 'find_sal').returns(sal).at_least(0)
        self.expect(f, 'name').returns('main').at_least(0)
        self.expect(f, 'pc').returns(0x1).at_least(0)
        self.expect(f, 'architecture').returns(architecture).at_least(0)
        self.expect(f, 'read_register').args(rax).returns(rax_value).at_least(0)
        self.expect(f, 'read_register').args(ymm0).times(0) # never
        self.expect(f, 'type').returns(0).at_least(0) # NORMAL_FRAME
        return f

    def test_sentry_variables(self):
        variables = preamble.SentryVariables(self.frame())
        self.assert_equal({'i': '123'}, variables.to_dict())

    def test_sentry_frame(self):
        sentry_frame = preamble.SentryFrame(self.frame())
        self.assert_equal({'filename': '$HOME/foo.so',
                           'function': 'main',
                           'instruction_addr': '0x1',
                           'lineno': 1, # NOTE: sentry starts at 1, gdb at 0
                           'package': '$HOME/foo.so',
                           'vars': {'i': '123'}}, sentry_frame.to_dict())

    def test_sentry_registers(self):
        registers = preamble.SentryRegisters(self.frame())
        self.assert_equal({'rax': '0x2'}, registers.to_dict())

    def test_sentry_thread(self):
        the_frame = self.frame()

        thread = self.mock()
        self.expect(thread.switch)
        thread.ptid = [123, 456, 789]
        thread.name = "TheThread"
        self.expect(thread.is_exited).returns(False)

        self.mock(gdb, 'newest_frame')
        self.expect(gdb.newest_frame).returns(the_frame)
        gdb.SIGTRAMP_FRAME = 4

        self.mock(gdb, 'FrameIterator')
        self.mock(gdb, 'FrameIterator.FrameIterator')
        self.expect(gdb.FrameIterator.FrameIterator).args(the_frame).returns([the_frame])

        thread = preamble.SentryThread(thread, is_crashed=False)
        self.assert_equal({'crashed': False,
                           'current': False,
                           'held_locks': {},
                           'state': None,
                            'id': 456,
                            'main': False,
                            'name': 'TheThread',
                            'stacktrace': {'frames': [{'filename': '$HOME/foo.so',
                                                        'function': 'main',
                                                        'instruction_addr': '0x1',
                                                        'lineno': 1, # NOTE: sentry starts at 1, gdb at 0
                                                        'package': '$HOME/foo.so',
                                                        'vars': {'i': '123'}}],
                                            'registers': {'rax': '0x2'}}}, thread.to_dict())

    def test_sentry_images(self):
        proc_mappings = """
process 207700
Mapped address spaces:

          Start Addr           End Addr       Size     Offset  Perms  objfile
      0x555555554000     0x555555556000     0x2000        0x0  r--p   /usr/bin/kwrite
      0x555555556000     0x555555558000     0x2000     0x2000  r-xp   /usr/bin/kwrite
      0x555555558000     0x555555559000     0x1000     0x4000  r--p   /usr/bin/kwrite
      0x55555555a000     0x55555555b000     0x1000     0x5000  r--p   /usr/bin/kwrite
      0x55555555b000     0x55555555c000     0x1000     0x6000  rw-p   /usr/bin/kwrite
      0x55555555c000     0x55555637e000   0xe22000        0x0  rw-p   [heap]
      0x7fffc0000000     0x7fffc0021000    0x21000        0x0  rw-p
      0x7fffc0021000     0x7fffc4000000  0x3fdf000        0x0  ---p
      0x7fffc7000000     0x7fffc703c000    0x3c000        0x0  r--p   /usr/lib/x86_64-linux-gnu/libx265.so.199
      0x7fffc703c000     0x7fffc7ed4000   0xe98000    0x3c000  r-xp   /usr/lib/x86_64-linux-gnu/libx265.so.199
      0x7fffc7ed4000     0x7fffc7f4f000    0x7b000   0xed4000  r--p   /usr/lib/x86_64-linux-gnu/libx265.so.199
      0x7fffc7f4f000     0x7fffc7f52000     0x3000   0xf4e000  r--p   /usr/lib/x86_64-linux-gnu/libx265.so.199
      0x7fffc7f52000     0x7fffc7f55000     0x3000   0xf51000  rw-p   /usr/lib/x86_64-linux-gnu/libx265.so.199
      0x7fffd5ff6000     0x7fffd5ffc000     0x6000        0x0  r--s   /var/cache/fontconfig/68a526f6-7ccc-4a32-af00-647cd10884c1-le64.cache-7
      0x7fffdc000000     0x7fffdc001000     0x1000        0x0  r--s   /home/me/.local/share/mime/mime.cache
      0x7fffdc476000     0x7fffdc676000   0x200000 0x118ded000  rw-s   /dev/dri/renderD128
      0x7ffff6200000     0x7ffff6228000    0x28000        0x0  r--p   /usr/lib/x86_64-linux-gnu/libc.so.6
      0x7ffff6228000     0x7ffff63bd000   0x195000    0x28000  r-xp   /usr/lib/x86_64-linux-gnu/libc.so.6
      0x7ffff63bd000     0x7ffff6415000    0x58000   0x1bd000  r--p   /usr/lib/x86_64-linux-gnu/libc.so.6
      0x7ffff6415000     0x7ffff6419000     0x4000   0x214000  r--p   /usr/lib/x86_64-linux-gnu/libc.so.6
      0x7ffff6419000     0x7ffff641b000     0x2000   0x218000  rw-p   /usr/lib/x86_64-linux-gnu/libc.so.6
      0x7ffff7fbd000     0x7ffff7fc1000     0x4000        0x0  r--p   [vvar]
      0x7ffff7fc1000     0x7ffff7fc3000     0x2000        0x0  r-xp   [vdso]
      0x7ffff7fc3000     0x7ffff7fc5000     0x2000        0x0  r--p   /usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2
      0x7ffff7fc5000     0x7ffff7fef000    0x2a000     0x2000  r-xp   /usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2
      0x7ffff7fef000     0x7ffff7ffa000     0xb000    0x2c000  r--p   /usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2
      0x7ffff7ffa000     0x7ffff7ffb000     0x1000 0x10663c000  rw-s   /dev/dri/renderD128
      0x7ffff7ffb000     0x7ffff7ffd000     0x2000    0x37000  r--p   /usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2
      0x7ffff7ffd000     0x7ffff7fff000     0x2000    0x39000  rw-p   /usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2
      0x7ffffffdd000     0x7ffffffff000    0x22000        0x0  rw-p   [stack]
  0xffffffffff600000 0xffffffffff601000     0x1000        0x0  --xp   [vsyscall]
        """

        self.mock(gdb, 'execute')
        self.expect(gdb.execute).args('info proc mappings', to_string=True).returns(proc_mappings)

        obj1 = self.mock()
        obj1.filename = '/usr/lib/x86_64-linux-gnu/libc.so.6'
        obj1.build_id = '161dd025ad435f0e873aafd55dc65d8a4cb1d93f'

        obj2 = self.mock()
        obj2.filename = '/usr/lib/x86_64-linux-gnu/libx265.so.199'
        obj2.build_id = '272ee025ad435f0e873aafd55dc65d8a4cb1d93f'

        self.mock(gdb, 'objfiles')
        self.expect(gdb.objfiles).returns([obj1, obj2])

        images = preamble.SentryImages()
        self.assert_equal([{'arch': 'x86_64',
            'code_file': '/usr/lib/x86_64-linux-gnu/libx265.so.199',
            'code_id': '272ee025ad435f0e873aafd55dc65d8a4cb1d93f',
            'debug_id': '25e02e27-43ad-0e5f-873a-afd55dc65d8a',
            'image_addr': '0x7fffc7000000',
            'image_size': 16076800,
            'type': 'elf'},
            {'arch': 'x86_64',
            'code_file': '/usr/lib/x86_64-linux-gnu/libc.so.6',
            'code_id': '161dd025ad435f0e873aafd55dc65d8a4cb1d93f',
            'debug_id': '25d01d16-43ad-0e5f-873a-afd55dc65d8a',
            'image_addr': '0x7ffff6200000',
            'image_size': 2207744,
            'type': 'elf'}]
            , images.to_list())

    def test_sentry_qml(self):
        thread = self.mock()
        self.expect(thread.is_valid).returns(True)
        self.expect(thread.switch)

        inferior = self.mock()
        self.expect(inferior.threads).returns([thread])

        symbol0 = self.mock()
        symbol0.is_variable = False
        symbol0.is_argument = False

        frame0 = self.mock(MagicMock)

        dereferenced_type = self.mock()
        dereferenced_type.name = 'QV4::ExecutionEngine'

        typeobj_target = self.mock()
        self.expect(typeobj_target.unqualified).returns(dereferenced_type)

        typeobj = self.mock()
        typeobj.code = 1
        self.expect(typeobj.target).returns(typeobj_target)

        value1 = MagicMock()
        value1.is_optimized_out = False
        value1.type = typeobj
        value1.__int__.return_value = 0x3

        symbol1 = self.mock()
        symbol1.is_variable = True
        symbol1.is_argument = False
        self.expect(symbol1.value).args(frame0).returns(value1)

        self.expect(frame0.__len__).returns(5)
        self.expect(frame0.block).returns([symbol0, symbol1])

        frame1 = self.mock(MagicMock)
        self.expect(frame1.__len__).returns(5)
        self.expect(frame1.older).returns(frame0)

        self.mock(gdb, 'selected_inferior')
        self.expect(gdb.selected_inferior).returns(inferior)
        self.mock(gdb, 'selected_thread')
        self.expect(gdb.selected_thread).returns(thread)
        self.mock(gdb, 'newest_frame')
        self.expect(gdb.newest_frame).returns(frame1)
        self.mock(gdb, 'parse_and_eval')
        qml_trace = '''
"stack=[frame={level=\\"0\\",func=\\"expression for onCompleted\\",file=\\"qrc:/ui/main.qml\\",fullname=\\"qrc:/ui/main.qml\\",line=\\"10\\",language=\\"js\\"}]"
        '''
        self.expect(gdb.parse_and_eval).args('qt_v4StackTraceForEngine((void*)0x3)').returns(qml_trace)
        gdb.TYPE_CODE_PTR = 1

        thread = preamble.SentryQMLThread()
        self.assert_equal({'crashed': True,
            'id': 'QML',
            'name': 'QML',
            'stacktrace': {'frames': [{'filename': 'qrc:/ui/main.qml',
            'function': 'expression for onCompleted',
            'in_app': True,
            'lineno': 10,
            'platform': 'other'}]}}, thread.to_dict())

    def test_sentry_image(self):
        objfile = self.mock()
        objfile.filename = '/usr/bin/true'
        objfile.build_id = '272ee025ad435f0e873aafd55dc65d8a4cb1d93f'

        self.mock(gdb, 'objfiles')
        self.expect(gdb.objfiles).returns([objfile])

        image = preamble.SentryImage('/usr/bin/true', 1000, 2000)
        self.assert_equal({'arch': 'x86_64',
            'code_file': '/usr/bin/true',
            'code_id': '272ee025ad435f0e873aafd55dc65d8a4cb1d93f',
            'debug_id': '25e02e27-43ad-0e5f-873a-afd55dc65d8a',
            'image_addr': '0x3e8',
            'image_size': 1000,
            'type': 'elf'}, image.to_dict())

    def test_sentry_image_mapping_fail(self):
        try:
            import sentry_sdk
        except ModuleNotFoundError:
            self.skipTest('sentry_sdk not available')

        objfile = self.mock()
        objfile.filename = '/usr/bin/true'
        objfile.build_id = '272ee025ad435f0e873aafd55dc65d8a4cb1d93f'

        self.mock(gdb, 'objfiles')
        self.expect(gdb.objfiles).returns([objfile])

        self.mock(gdb, 'lookup_objfile')
        self.expect(gdb.lookup_objfile).returns(None)

        progspace = self.mock()
        inferior = self.mock()
        inferior.pid = 0
        self.expect(progspace.is_valid).returns(False)
        inferior.progspace = progspace
        self.mock(gdb, 'selected_inferior')
        self.expect(gdb.selected_inferior).returns(inferior)

        self.mock(sentry_sdk, 'capture_exception')
        self.expect(sentry_sdk.capture_exception)
        preamble.SentryImage('/usr/bin/true.so', 1000, 2000)

    def test_print_qml_frames(self):
        # the payload is missing a line -> we expect no assertions of any kind!
        preamble.print_qml_frames('frame={level="0",func="saveConfig",file="/data/projects/kde/usr/share/plasma/shells/org.kde.plasma.desktop/contents/configuration/ConfigurationContainmentAppearance.qml",fullname="/data/project".')

if __name__ == '__main__':
    unittest.main()

